from collections import deque
import dataclasses
import threading
from typing import Type
from functools import wraps
import multiprocessing as mp
from multiprocessing.managers import BaseProxy, SyncManager
import multiprocessing.managers
import rich
from rich import progress
from rich.progress import TaskID
from .executor import T_POOL_MODE
import mqdm as M


class Task(progress.Task):
    @property
    def speed(self):
        """Optional[float]: Get the estimated speed in steps per second."""
        if self.start_time is None:
            return None
        with self._lock:
            progress = self._progress
            if progress:
                total_time = progress[-1].timestamp - self.start_time
            elif self.finished:
                total_time = self.finished_time
            else:
                total_time = self.elapsed
            return self.completed / total_time if total_time else total_time

            # total_time = progress[-1].timestamp - progress[0].timestamp
            # if total_time == 0:
            #     return None
            # iter_progress = iter(progress)
            # next(iter_progress)
            # total_completed = sum(sample.completed for sample in iter_progress)
            # speed = total_completed / total_time
            # # if self.id == 0:
            # #     print(progress)
            # #     print("speed p", speed, total_completed, total_time)
            # return speed

    # def reset(self):
    #     super().reset()
    #     self._progress.append(progress.ProgressSample(self.start_time, self.completed))
    #     # self._progress.append(progress.ProgressSample(self.get_time(), self.completed))


class Progress(progress.Progress):
    multiprocess = False

    def __init__(self, *columns, _tasks=None, _task_index=None, _pause_event=None, **kw):
        super().__init__(*columns, **kw)

        # save init options in case we need to recreate the object in a different process
        self._init_options = kw
        # load serialized tasks from another progress object
        if _tasks is not None:
            self._tasks = {task_id: self._load_task(**task) for task_id, task in _tasks.items()}
        self._task_index = progress.TaskID(_task_index or 0)
        # self._pause_event = _pause_event
        # if self._pause_event is not None:  # is this necessary if start exists?
        #     self._pause_event.set()

    def print(self, *args, **kw):
        """Print to the console."""
        rich.print(*args, **kw)

    @wraps(progress.Progress.update)
    def update(self, task_id, **kw):
        if 'description' in kw and kw['description'] is None:  # ignore None descriptions
            kw.pop('description')
        return super().update(task_id, **kw)
    
    @wraps(progress.Progress.update)
    def update_(self, task_id, **kw):
        try:
            return self.update(task_id, **kw)
        except KeyError as e:
            pass

    def add_task(
        self,
        description: str='',
        start: bool = True,
        total: float|None = 100.0,
        completed: int = 0,
        visible: bool = True,
        **fields,
    ) -> progress.TaskID:
        with self._lock:
            task = self._new_task(description or '', start, total, completed, visible, **fields)
            self._tasks[task.id] = task
        self.refresh()
        return task.id

    def start_task(self, task_id: progress.TaskID) -> None:
        with self._lock:
            self._start_task(self._tasks[task_id])

    # def stop_task(self, task_id):
    #     return super().stop_task(task_id)

    def clear(self):
        """Clear the progress bar."""
        with self._lock:
            self._tasks.clear()
    
    # def start(self):
    #     if self._pause_event is not None:
    #         self._pause_event.set()
    #     super().start()
    
    # def stop(self):
    #     if self._pause_event is not None:
    #         self._pause_event.clear()
    #     super().stop()

    def pop_task(self, task_id, remove=None):
        """Close a task and return its serialized data."""
        try:
            self.stop_task(task_id)
            data = self.dump_task(task_id)
            if remove is None:
                remove = self._tasks[task_id].fields.get('transient', False)
            if remove:
                self.remove_task(task_id)
            return data
        except KeyError as e:
            pass

    # convert to multiprocessing proxy

    def convert_proxy(self) -> 'ProgressProxy':
        """Convert to a multiprocessing proxy object so methods can be called in another process."""
        # get current state and cleanup
        started = self.live.is_started
        tasks = self.dump_tasks()
        for task_id in tasks:
            self.remove_task(task_id)
        started or self.start()
        self.refresh()
        self.stop()

        # create proxy
        proxy = get_manager().mqdm_Progress(
            *self.columns,
            _tasks=tasks,
            _task_index=self._task_index,
            **self._init_options,
        )
        proxy.multiprocess = True
        if started:
            proxy.start()
        return proxy
    
    # def get_task_attr(self, task_id, attr, default=None):
    #     task = self._tasks[task_id]
    #     try:
    #         return getattr(task, attr)
    #     except AttributeError:
    #         return task.fields.get(attr, default)

    def dump_tasks(self):
        with self._lock:
            return {task_id: self._dump_task(self._tasks[task_id]) for task_id in self._tasks}

    def dump_task(self, task_id):
        with self._lock:
            return self._dump_task(self._tasks[task_id])

    def _dump_task(self, task):
        d = {k.name: getattr(task, k.name) for k in dataclasses.fields(task) if not k.name.startswith('_')}
        d['_progress'] = []
        if task._progress:
            d['_progress'].append((task._progress[-1].timestamp, sum(s.completed for s in task._progress)))
        return d
    
    def _load_task(self, *, start_time=None, stop_time=None, start=True, _progress=None, **data):
        task = Task(_get_time=self.get_time, _lock=self._lock, **data)
        task.start_time = start_time
        task.stop_time = stop_time
        start and self._start_task(task)
        if _progress:
            task._progress = deque([progress.ProgressSample(*s) for s in _progress], maxlen=1000)
        return task
    
    def _start_task(self, task):
        if task.start_time is None:
            task.start_time = self.get_time()
        # if task.start_time is not None and not task._progress:
        #     task._progress.append(progress.ProgressSample(task.start_time, task.completed))

    def load_task(self, task: dict, start=True):
        with self._lock:
            task = self._load_task(start=start, **task)
            self._tasks[task.id] = task
            if task.id >= self._task_index:
                self._task_index = progress.TaskID(task.id+1)
        self.refresh()

    def _new_task(self, 
                  description: str='',
                  start: bool = True,
                  total: float|None = None,
                  completed: int = 0,
                  visible: bool = True,
                  **fields):
        task = Task(
            self._task_index,
            description,
            total,
            completed,
            visible=visible,
            fields=fields,
            _get_time=self.get_time,
            _lock=self._lock,
        )
        start and self._start_task(task)
        # task._progress.append(progress.ProgressSample(task.start_time or self.get_time(), task.completed))
        self._task_index = progress.TaskID(int(self._task_index) + 1)
        return task

    def new_task(self,
                  description: str='',
                  start: bool = True,
                  total: float|None = None,
                  completed: int = 0,
                  visible: bool = True,
                  **fields):
        with self._lock:
            return self._dump_task(self._new_task(description or '', start, total, completed, visible, **fields))



def proxymethod(func):
    name = func.__name__
    @wraps(func)
    def _call(self, *a, **kw):
        return self._callmethod(name, a, kw)
    _call._is_exposed_ = True
    return _call


class ProgressProxy(BaseProxy):
    multiprocess = True
    # def __init__(self, *args, **kw):
    #     console = rich.get_console()
    #     self._redirect_io = _RedirectIO(console)
    #     self._redirect_io.enable()
    #     super().__init__(*args, **kw)

    start_task = proxymethod(Progress.start_task)
    stop_task = proxymethod(Progress.stop_task)
    add_task = proxymethod(Progress.add_task)
    remove_task = proxymethod(Progress.remove_task)
    update = proxymethod(Progress.update)
    update_ = proxymethod(Progress.update_)
    refresh = proxymethod(Progress.refresh)
    start = proxymethod(Progress.start)
    stop = proxymethod(Progress.stop)
    print = proxymethod(Progress.print)
    # get_task_attr = proxymethod(Progress.get_task_attr)
    dump_task = proxymethod(Progress.dump_task)
    dump_tasks = proxymethod(Progress.dump_tasks)
    load_task = proxymethod(Progress.load_task)
    new_task = proxymethod(Progress.new_task)
    pop_task = proxymethod(Progress.pop_task)
    clear = proxymethod(Progress.clear)

ProgressProxy._exposed_ = tuple(k for k, v in ProgressProxy.__dict__.items() if getattr(v, '_is_exposed_', False))

multiprocessing.managers.ProgressProxy = ProgressProxy  # Can't pickle - attribute lookup ProgressProxy on multiprocessing.managers failed


class MqdmManager(SyncManager):
    mqdm_Progress: Type[ProgressProxy]
MqdmManager.register('mqdm_Progress', Progress, ProgressProxy)

M._manager = None
M._pause_event = threading.Event()
M._pause_event.set()
M._shutdown_event = threading.Event()
M._shutdown_event.set()
def get_manager() -> MqdmManager:
    if M._manager is not None:
        return M._manager
    M._manager = manager = MqdmManager()
    manager.start()
    M._pause_event = manager.Event()
    M._pause_event.set()
    M._shutdown_event = manager.Event()
    M._shutdown_event.set()
    return manager

import atexit

def _shutdown_manager():
    try:
        manager = getattr(M, '_manager', None)
        if manager is not None:
            manager.shutdown()
            M._manager = None
    except Exception:
        # Best-effort shutdown; ignore errors at interpreter teardown
        pass

atexit.register(_shutdown_manager)


def get_progress_instance(pool_mode: T_POOL_MODE=None, *columns, **kw):
    if pool_mode == 'process':
        manager = get_manager()
        return manager.mqdm_Progress(*columns, **kw)
    return Progress(*columns, **kw)



# import sys
# import rich
# from typing import cast, TextIO, IO, Optional
# from rich.file_proxy import FileProxy
# class _RedirectIO:
#     def __init__(self, console: rich.console.Console, redirect_stdout: bool = True, redirect_stderr: bool = True):
#         self.console_compatible = console.is_terminal or console.is_jupyter
#         self._redirect_stdout = redirect_stdout
#         self._redirect_stderr = redirect_stderr
#         self._restore_stdout: Optional[IO[str]] = None
#         self._restore_stderr: Optional[IO[str]] = None

#     def __enter__(self):
#         self.enable()
#         return self
    
#     def __exit__(self, *exc):
#         self.disable()

#     def __del__(self):
#         self.disable()

#     def enable(self) -> None:
#         """Enable redirecting of stdout / stderr."""
#         if self.console_compatible:
#             if self._redirect_stdout and not isinstance(sys.stdout, FileProxy):
#                 self._restore_stdout = sys.stdout
#                 sys.stdout = cast("TextIO", FileProxy(self.console, sys.stdout))
#             if self._redirect_stderr and not isinstance(sys.stderr, FileProxy):
#                 self._restore_stderr = sys.stderr
#                 sys.stderr = cast("TextIO", FileProxy(self.console, sys.stderr))

#     def disable(self) -> None:
#         """Disable redirecting of stdout / stderr."""
#         if self._restore_stdout:
#             sys.stdout = cast("TextIO", self._restore_stdout)
#             self._restore_stdout = None
#         if self._restore_stderr:
#             sys.stderr = cast("TextIO", self._restore_stderr)
#             self._restore_stderr = None



import time
# import multiprocessing as mp
# mp.set_start_method('spawn')
# mp.set_start_method('fork')
# mp.set_start_method('forkserver')
import mqdm
from mqdm import examples

@mqdm.iex
def main():
    _rich_traceback_omit = True
    import fire
    fire.Fire({
        # 'bars': bar.example_pool,
        # 'bar': bar.example_bar,
        # 'fn': bar.example_fn,
        # 'messy': bar.example_messy,
        k.removeprefix('example_'): v for k, v in examples.__dict__.items() if k.startswith('example_')
    })
main()


import time
from contextlib import contextmanager

import rich
from rich import progress


import mqdm as M  # self
from . import proxy
from . import utils
# from .executor import SequentialExecutor, ProcessPoolExecutor, ThreadPoolExecutor, Executor
from .executor import executor, T_POOL_MODE, Initializer
from ._dev import embed, inp, bp, iex, profile, timeit
from .utils import args, fopen, ratelimit
M.input = inp


_manager: 'proxy.MqdmManager' = None
_instances: 'list[M.mqdm]' = []
_keep = False
pbar: 'proxy.Progress|proxy.ProgressProxy' = None


# ---------------------------------------------------------------------------- #
#                               Progress methods                               #
# ---------------------------------------------------------------------------- #


def _new_pbar(pool_mode: T_POOL_MODE=None, bytes=False, **kw) -> 'proxy.Progress|proxy.ProgressProxy':
    kw.setdefault('refresh_per_second', 8)
    return proxy.get_progress_instance(
        pool_mode,
        "[progress.description]{task.description}",
        progress.BarColumn(bar_width=None),
        "[progress.percentage]{task.percentage:>3.0f}%",
        utils.MofNColumn(bytes=bytes),
        utils.SpeedColumn(bytes=bytes),
        utils.TimeElapsedColumn(compact=True),
        progress.TimeRemainingColumn(compact=True),
        progress.SpinnerColumn(),
        **kw,
    )


def _get_pbar(pool_mode: T_POOL_MODE=None, start=True, **kw) -> 'proxy.Progress|proxy.ProgressProxy':
    # no progress bar yet, create one
    if not M.pbar:
        # print("New progress bar", pool_mode)
        M.pbar = _new_pbar(pool_mode=pool_mode, **kw)
    # need to create multiprocess-compatible progress bar
    elif pool_mode == 'process' and not M.pbar.multiprocess:
        # print("Converting proxy")
        M.pbar = M.pbar.convert_proxy()
    if start:
        M.pbar.start()
    return M.pbar


def _clear_pbar(strict=True, force=False, soft=False):
    """Clear the progress bar."""
    if force:
        for bar in _instances[::-1]:
            bar._remove(False)
            bar.disable = True
        M.pbar.stop()
        M.pbar = None
    if M._instances:
        if strict:
            raise RuntimeError("Cannot clear progress bar while instances are still active.")
    elif not utils.is_main_process():
        if strict:
            raise RuntimeError("Cannot clear progress bar in a subprocess.")
    elif soft or M._keep:
        if M.pbar is not None:
            M.pbar.refresh()
    else:
        if M.pbar is not None:
            M.pbar.start()
            M.pbar.refresh()
            M.pbar.stop()
        M.pbar = None


@contextmanager
def group():
    """Group progress bars."""
    try:
        M._keep = True
        yield 
    finally:
        M._keep = False
        _clear_pbar()


# ---------------------------------------------------------------------------- #
#                            Global context methods                            #
# ---------------------------------------------------------------------------- #


def print(*args, **kw):
    """Print with rich."""
    if pbar is not None:
        return pbar.print(*args, **kw)
    return rich.print(*args, **kw)


def get(i=-1):
    """Get an mqdm instance."""
    try:
        return list(_instances.values())[i]()
    except IndexError:
        raise IndexError(f'No progress bar found at index {i} in list of length {len(_instances)}')


def set_description(desc, i=-1):
    """Set the description of the last progress bar."""
    return get(i).set_description(desc)


def set(i=-1, **kw):
    """Set the last progress bar."""
    return get(i).set(**kw)


def update(n=1, i=-1, **kw):
    """Update the last progress bar."""
    return get(i).update(n, **kw)



import weakref
from collections import OrderedDict

_instances = OrderedDict()

def _add_instance(bar):
    if bar not in _instances:
        _instances[hash(bar)] = weakref.ref(bar)
    return bar

def _remove_instance(bar):
    _instances.pop(hash(bar), None)  # Safely remove without error if bar is not found

def _close_instances():
    for bar_ref in list(_instances.values()):  # Make sure to use list() to avoid modifying while iterating
        bar = bar_ref()
        if bar is not None:
            bar.close()
    _instances.clear()

import atexit
atexit.register(_close_instances)


def _pause_wait():
    M._pause_event.wait()

_last_pause_wait_time = 0
_pause_wait_ttl_seconds = 0.2
def _ttl_pause_wait(): # lru_cache(1)(_pause_wait)(int(time.time()/ttl))
    global _last_pause_wait_time
    current_time = int(time.time() / _pause_wait_ttl_seconds)
    if current_time != _last_pause_wait_time:
        _last_pause_wait_time = current_time
        M._pause_event.wait()


def pause(paused=True):
    """Pause the progress bars. Useful for opening an interactive shell or printing stack traces."""
    prev_paused = getattr(pbar, 'paused', False)
    if pbar is not None:
        pbar.paused = paused
        if paused:
            pbar.stop()
            M._pause_event.clear()
        else:
            pbar.start()
            M._pause_event.set()
    return _pause_exit(prev_paused)

class _pause_exit:
    def __init__(self, prev_paused):
        self.prev_paused = prev_paused  # it was paused before we got here
        _pause_exit.last = self  # if another pause was called, ignore this one
    def __enter__(self): pass
    def __exit__(self, c, exc, t): 
        if not exc and not self.prev_paused and self is _pause_exit.last:  # dont unpause for exceptions
            pause(False)


# ---------------------------------------------------------------------------- #
#                               Primary interface                              #
# ---------------------------------------------------------------------------- #


from .bar import mqdm, pool, ipool

# more descriptive names to avoid polluting the namespace
mqpool = pool
mqipool = ipool

__all__ = [
    'mqdm',
    'mqpool',
    'mqipool',
]


import os
import time
import multiprocessing as mp
from rich import progress
import mqdm as M

# ---------------------------------------------------------------------------- #
#                                     Utils                                    #
# ---------------------------------------------------------------------------- #


def is_main_process():
    """Check if the current process is the main process."""
    return mp.current_process().name == 'MainProcess'

def process_name():
    """Get the name of the current process."""
    return mp.current_process().name

class args:
    '''Storing Function Arguments for later.
    
    Example:
    ```
    def fn(a, b=2, c=3):
        print(a, b, c)

    fn_args = [args(i, c=i*2) for i in range(3)]
    for arg in fn_args:
        arg(fn, b=2)
    ```
    '''
    def __init__(self, *a, _MQDM_PARENT_ARGS_OBJ_=None, **kw):
        if _MQDM_PARENT_ARGS_OBJ_ is not None:
            a = _MQDM_PARENT_ARGS_OBJ_.a + a
            kw = {**_MQDM_PARENT_ARGS_OBJ_.kw, **kw}

        self.a = a
        self.kw = kw

    def __repr__(self):
        args=', '.join([f'\n  {x!r}' for x in self.a] + [f'\n  {k}={v!r}' for k, v in self.kw.items()])
        args = args + '\n' if args else ''
        return f"({args})"

    def __getitem__(self, i):
        return self.a[i] if isinstance(i, int) else self.kw[i]

    def __call__(self, fn, *a, **kw):
        return fn(*self.a, *a, **dict(self.kw, **kw))

    @classmethod
    def from_item(cls, x, *a, **kw):
        return cls(*a, _MQDM_PARENT_ARGS_OBJ_=x, **kw) if isinstance(x, cls) else cls(x, *a, **kw)

    @classmethod
    def from_items(cls, items, *a, **kw):
        return [cls.from_item(x, *a, **kw) for x in items]
    
    @classmethod
    def from_tuples(cls, items, *a, **kw):
        return [cls.from_item(*x, *a, **kw) for x in items]


def maybe_call(fn, *a, **kw):
    """Call the value if it is callable. Otherwise, return it."""
    return fn(*a, **kw) if callable(fn) else fn


def try_len(it, default=None):
    """Try to get the length of an object, returning a default value if it fails."""
    if it is None:
        return default
    if isinstance(it, int):
        return it
    try:
        return len(it)
    except TypeError:
        pass

    try:
        x = type(it).__length_hint__(it)
        return x if isinstance(x, int) else default
    except (AttributeError, TypeError):
        return default


# ---------------------------------------------------------------------------- #
#                         Custom Progress Column Types                         #
# ---------------------------------------------------------------------------- #


class MofNColumn(progress.DownloadColumn):
    '''A progress column that shows the current vs. total count of items.'''
    def __init__(self, bytes=False, separator="/", **kw):
        self.bytes = bytes
        self.separator = separator
        super().__init__(**kw)

    def render(self, task):
        bytes = task.fields.get("bytes", self.bytes)
        if bytes:
            return super().render(task)
        total = f'{int(task.total):,}' if task.total is not None else "?"
        return progress.Text(
            f"{int(task.completed):,d}{self.separator}{total}",
            style="progress.download",
            justify='right'
        )


class SpeedColumn(progress.TransferSpeedColumn):
    """Renders human readable transfer speed."""
    def __init__(self, bytes=False, unit_scale=1, **kw):
        self.bytes = bytes
        self.unit_scale = unit_scale
        super().__init__(**kw)

    def render(self, task):
        """Show data transfer speed."""
        bytes = task.fields.get("bytes", self.bytes)
        if bytes:
            return super().render(task)
        speed = task.finished_speed or task.speed
        if speed is None:
            return progress.Text(f"{speed}", style="progress.data.speed")
        end = '/s'
        if 0 < speed < 1:
            speed = 1 / speed
            speed, suffix = time_units(speed)
            end = '/x'
            return progress.Text(f"{speed:.1f}{suffix}{end}", justify='right', style="progress.data.speed")
        unit, suffix = progress.filesize.pick_unit_and_suffix(
            int(speed), 
            ["x", "K", "M", "B", "T"], 
            # ["", "×10³", "×10⁶", "×10⁹", "×10¹²"], 
            1000)
        return progress.Text(f"{speed/unit:.1f}{suffix}{end}", justify='right', style="progress.data.speed")

def time_units(seconds):
    for d, u in [(60, 's'), (60, 'm'), (24, 'h'), (365, 'd')]:
        if seconds < d:
            return seconds, u
        seconds /= d
    return seconds, 'y'


class TimeElapsedColumn(progress.TimeRemainingColumn):
    """Renders time elapsed."""
    def __init__(self, compact: bool = False, **kw):
        self._compact = compact
        super().__init__(**kw)

    def render(self, task):
        """Show time elapsed."""
        elapsed = task.finished_time if task.finished else task.elapsed
        if elapsed is None:
            return progress.Text("--:--" if self._compact else "-:--:--", style="progress.elapsed")
        minutes, seconds = divmod(int(elapsed), 60)
        hours, minutes = divmod(minutes, 60)
        return progress.Text(
            (f"{hours:d}:" if hours or not self._compact else "") + 
            f"{minutes:02d}:{seconds:02d}", style="progress.elapsed")



# class LogBarColumn(progress.BarColumn):
#     def render(self, task):
#         return progress.Group(
#             super().render(task),
#             progress.Text(task.description, style="progress.description"),
#         )



class fopen:
    """Open a file with a progress bar."""
    def __init__(self, fname: os.PathLike, mode='r', pbar: 'M.mqdm'=None, **kw):
        self.total = os.path.getsize(fname)
        self.fd = open(fname, mode)
        if pbar is None:
            kw.setdefault('desc', os.path.basename(fname))
            pbar = M.mqdm(init_kw={'bytes': True}, **kw)
        self.pbar = pbar
        self._pbar_managed = not pbar.entered
        pbar.update(total=self.total)
        
    def __enter__(self):
        self.fd.__enter__()
        if self._pbar_managed:
            self.pbar.__enter__()
        return self
    
    def __exit__(self, *args):
        self.fd.__exit__(*args)
        if self._pbar_managed:
            self.pbar.__exit__(*args)

    def __iter__(self):
        return self

    def __next__(self):
        line = next(self.fd)
        self.pbar.fast_advance(len(line))
        return line

    def __getattr__(self, name):
        return getattr(self.fd, name)
    
    def set(self, **kw):
        self.pbar.set(**kw)
        return self
    
    def set_description(self, desc):
        self.pbar.set_description(desc)
        return self


def ratelimit(iter, seconds):
    """Limit the rate of an iterator."""
    if seconds is None:
        return iter
    lag = 0
    t0 = time.time()
    for x in iter:
        yield x
        t = time.time()
        dt = seconds - (t - t0) - lag
        if dt > 0:
            time.sleep(dt)
        lag = max(-dt, 0)
        t0 = t + max(0, dt)



import setuptools

setuptools.setup(
    name='mqdm',
    version='1.2.0',
    description='cross-process progress bars',
    long_description=open('README.md').read().strip(),
    long_description_content_type='text/markdown',
    packages=setuptools.find_packages(),
    python_requires='>=3.10',
    install_requires=[
        'rich>=13'
    ],
    extras_require={
        'test': [
            'pytest',
            'fire'
        ],
        'dev': [
            'fire',
            'ipython',
            'pyinstrument',
            'pdbr'
        ]
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Environment :: Console',
        'Topic :: Software Development :: Libraries',
        'Topic :: Utilities',
    ],
)
